% Shape deformation using the bending energy and SVM regression as in Song
% Wang's paper.
%% Init and read all sep files
%  should call deform_init FIRST!

%% Use median filtering, not work well
I = envsI;
%Im = medfilt2(I,[5,5]);
Im = I;
%subplot(1,2,1), imshow(Im,[]),title('median filter'); subplot(1,2,2), imshow(I,[]), title('I');

%% Smooth the envelope using the aniso diffusion
num_iter = 5;
delta_t = 1/7;
kappa = 20;
option = 1;
envsaI = anisodiff2D(I,num_iter,delta_t,kappa,option);
% subplot 121, imshow(Im,[]), subplot 122, imshow(envsaI,[]), title('ad');


%% Then we need to determine the contour's normal direction, to determined 
% how the markers move.
% First find the normal direction for all marker pts. NOTICE that the
% contour pts is ordered in the clock-wise direction
mz = markerTc(:,1)'; mx = markerTc(:,2)';
dz = [mz(2:end) mz(1)] - [mz(end) mz(1:end-1)];
dx = [mx(2:end) mx(1)] - [mx(end) mx(1:end-1)];
amp = hypot(dz,dx); % calc sqrt(dz*dz+dx*dx)
dx = dx./amp; dz = dz./amp;
% tangent vector, rotate it by 90deg. we get the normal direction
tv = [dz; dx];
[th,r] = cart2pol(tv(2,:),tv(1,:));
th = th - pi*0.5;
nv = zeros(size(tv));
[nv(2,:) nv(1,:)] = pol2cart(th,r);

%% Might want to smooth it a little
nvs = zeros(size(nv)); span = 13;
nvs(2,:) = smooth(nv(2,:),span,'sgolay'); nvs(1,:) = smooth(nv(1,:),'lowess');
% normalize again.
A = hypot(nvs(1,:),nvs(2,:));
nvs = nvs ./ repmat(A,2,1);



%% Then for each landmark, find a set of possible corresponding landmark
% points Bi = {vi(j),j=1,2,...} in the input image.
nv = nvs;
% The most simply would be, the regional maxima on the direction of
% searching lines, nv, the aniso seems to help
L = 20.0; % serach range
thrDetect = 0.1; %only peak with amplitude bigger than this thres will be considered

nmark = size(markerTc,1);
B = cell(1,nmark); 
% each element of B is a npts X 3 matrix, which contains z,x,val for each
% candidate points.
xs = [0 0]; ys= [0 0];
nvv = nv(:,1);
for i=1:1:nmark
  %extract along this direction
  if (selmInd(i)==true); %otherwise dont't even try to search
  nvv = nv(:,i);
  xs =  [mx(i) - nvv(2)*L*0.5 mx(i)+nvv(2)*L*0.5]; 
  ys =  [mz(i) - nvv(1)*L*0.5 mz(i)+nvv(1)*L*0.5];
  [cx,cz,c] = improfile(envsaI,xs,ys,L*2,'bilinear');
  r  =  linspace(-.5*L, .5*L, length(c))'; %The distance of each point to the marker pt.
  
  %if out the coord is boundary, then it shows NaN, first take out NaN
  sel = ~(isnan(c)); cx = cx(sel); cz = cz(sel); c = c(sel);
  %Find all local maxima with thres bigger than
  [vals inds] = find1dLocalMax(c,thrDetect);
  % put them into B
  if ~isempty(inds)
    B{i} = [cz(inds) cx(inds) vals r(inds)]; %The fourth elements tells hi.
  end
  end
end

% % % Plot these candidates pts with the envelope, put to zero.
% % Icand = envsT; 
% % Icand = envsaI; 
% % vmax = max(envsaI(:));
% % markInd = 137;
% % 
% % Plot the original pt and the cand pts.
% % if ~isempty(B{markInd})
% %   Icand(mz(markInd):mz(markInd),mx(markInd)) = vmax;
% %   coord = B{markInd}(:,1:2); coord = int32(coord);
% %   Icand(sub2ind(dim,coord(:,1),coord(:,2))) = vmax;
% %   imshow(Icand,[]); title(['temp marker and and its cand pts at mark index: ' num2str(markInd)]); %fig4(a);
% %   bis = B{markInd};
% %   hold on; scatter(bis(:,2),bis(:,1),'r'); hold off;
% % end

%% First we try to see what kind of interpolation this thin plate kernel does
%  in terms of landmark mapping application, following the expressions on
%  page 3 of the shape deformation paper.

% First define the mapping markers, v and v_hat.
v = zeros(nmark,2); v_hat = zeros(nmark,2); %maximum it can be
vnew = zeros(nmark,2);
indv = 0; indvnew = 0;
selB = false(1,nmark);
% I choose to take the whole 60 markers out
nleaveout = 0;
dsr = 1; % Sample marker pt every dsr pts.
for i=1:1:nmark-nleaveout
  Bi = B{i};
  if (mod(i,dsr)==0 && ~isempty(B{i}))
    indv = indv + 1;
    v_hat(indv,:) = Bi(1,1:2);
    v(indv,:) = markerTc(i,:);
    selB(i) = true;
  else
    indvnew = indvnew + 1;
    vnew(indvnew,:) = markerTc(i,:);
  end
end

% All the rest nleaveout pts are considered missing.
vnew(indvnew+1:indvnew+nleaveout,:) = markerTc(nmark-nleaveout+1:nmark, :);
indvnew = indvnew + nleaveout;  
% Clip the zero entries.
v = v(1:indv,:); v_hat = v_hat(1:indv,:); vnew = vnew(1:indvnew,:);
n = size(v,1); 

%% Solve the interpolation coefficient
% Form the matrix K,P and solve the equation 6, given a set of marker pts.
% v holds the n by 2 marker points coord.,
v_hat; % is target mapping position of set V
K = thinPlateKernel(v,v); P = [ones(n,1) v];
Kb = [K P; P' zeros(3,3)]; 
cond(Kb);

%% Then solve for the answer
rhs = [v_hat; zeros(3,2)];
lhs = Kb\rhs;
c   = lhs(1:n,1);       d = lhs(1:n,2);
a   = lhs(n+1:end,1);   b = lhs(n+1:end,2);
%% Then do the mapping for a bunch of new pts 
nn = size(vnew,1);
%we need to re-assemble a new matrix K and P.
Kn = thinPlateKernel(vnew,v);
Pn = [ones(nn,1) vnew];
vnew_hat = [Kn Pn] * lhs;

%% arrange the sets of Bi for further usage.
% Select the initial set of V' landmarks, put them into the vectors h,
% where (x,y) + h * (nvx,nvy) = x^,y^.
n; % n is the number of non-empty markers.
candIndBg= ones(1,n); % start from the the first element of B1, globally
Bs = B(selB); %Pick the non-empty Bi s.
nB = zeros(1,n);    %store number of candidate pts in each Bi.


% convert the Bs into a large array.
matBs = cell2mat(Bs'); % matBs is a N by 4 double arrays.
for i=2:1:length(Bs)
  nBi = size(Bs{i-1},1);
  nB(i-1) = nBi;
  candIndBg(i) = candIndBg(i-1)+nBi;
end
nB(end) = size(Bs{end},1); %not necessary.

%% Enter the SVM regression phase.
Lm = inv(Kb); assert(sum(isnan(Lm(:)))==0);
Lm = Lm(1:n,1:n);
assert(n == size(v,1)); %number of non-empty landmark candidates
nvs = nv(:,selB);

Nx = diag(nvs(1,:)); Ny = diag(nvs(2,:)); % Here x->Z, y->X in our case
H = inv(Nx'*Lm*Nx + Ny'*Lm*Ny);
p = H*(Nx'*Lm*v(:,1) + Ny'*Lm*v(:,2)); %This p vector is extremely small because Lm*v -> 0, for some unknown reason.

%%
alpha = zeros(2*n,1); %First half holds vec alpha, second half holds vec alpha_hat.
mMinus = [eye(n) -1*eye(n)];
mPlus  = [eye(n)    eye(n)];

Hn = mMinus' * H * mMinus; 
Hn = (Hn+Hn')/2; % enforce the symmetry, because Hn is not exactly symmetric due to numeric error.
bupdated = false;
gamma = zeros(n,1); %Stores the final mapping landmarks, t(v)=v+gamma*n

%% Init parameter, re run this for many times
lamda = 10;
ep=L/2 * 0.8;
c = 4*pi/(n*lamda);
niter = 10; % should be >1
alpha_thres = 1e-7; %Test if alpha can be seen as zero
%c = 1;
candIndB = ones(1,n); % start from the first pt in each Bi, local

%%
t = 1:1:n;
it = 1;
while (ep >= 0 && it<niter+1) %should be .5

  % Then extract the h array first, notice vector h changes every iteration
  cind = candIndBg + candIndB - 1;
  h    = matBs(cind,4);
  epv = ep * ones(1,n);  
  fn = mMinus'*(p+h) + mPlus'*epv';

  % solve the quad programming problem
  [alphas fval exitflag] = quadprog(Hn,fn,[],[],[],[],0*ones(2*n,1),c*ones(2*n,1));
  % [alphas fval exitflag] = quadprog(Hn,fn,[],[],[],[],0,c);
  if (exitflag >=0) %converged
    disp(['Exit flag for quad programming is normal but not 0/1:' num2str(exitflag)]);
  else
    disp(['!!Exit flag for quad programming is abnormal:' num2str(exitflag)]);
    %exit(-1);
  end
  % Then find gamma using eq (9).
  gamma = -H*mMinus*alphas - p;
  
  bupdated = false;
  
  %Find the support vectors, basically alpha or alpha^ is nonzero
  sel_supp = (alphas(1:n) > alpha_thres ) | ( alphas(n+1:end) > alpha_thres);
  % show the index of the support vector pts 
  suppvec_ind = find(sel_supp==1);  
  gh = (gamma - h)'; gh(sel_supp);
  %do we have candidate left to update V'
  minInd = 0;

  for j = suppvec_ind' %tranverse all support vec pts.
    lb = candIndBg(j); ub = lb + nB(j) - 1;
    hh = matBs(candIndBg(j):candIndBg(j)+nB(j)-1,4); % see if we can find a better candidate
    gamma_minus_h = abs(hh - gamma(j)); % Find the smallest.
    minInd = find(gamma_minus_h == min(gamma_minus_h(:)));
    if (minInd(1) ~= candIndB(j)) % we can update in this case
      candIndB(j) = minInd(1);
      bupdated = true;
    end
  end

  % Let's plot the gamma and h, plus the sel_supp index. 
  %Plot the updated points
  cind_old = cind; %update the cind index.
  cind = candIndBg + candIndB - 1;
  h    = matBs(cind,4); %recalculate h

  if (~bupdated) %shrink the epsilon
    ep = ep * 0.5;
    if (ep==0); break; end;
    if (ep>0 && ep<1); ep=0; end;
  else  
  end

it = it + 1; %the total number of iteration
end %while

%% Adjust the lamda parameter, we invert for the contour
lamda = 50;
c = 4*pi/(n*lamda);

% solve the quad programming problem
[alphas fval exitflag] = quadprog(Hn,fn,[],[],[],[],0*ones(2*n,1),c*ones(2*n,1));
% [alphas fval exitflag] = quadprog(Hn,fn,[],[],[],[],0,c);
if (exitflag >=0) %converged
  disp(['Exit flag for quad programming is normal but not 0/1:' num2str(exitflag)]);
else
  disp(['!!Exit flag for quad programming is abnormal:' num2str(exitflag)]);
  %exit(-1);
end
% Then find gamma using eq (9).
gamma = -H*mMinus*alphas - p;

% Plot the reglarized mapping output
v_hc = v + nvs'.* repmat(gamma,1,2);
tpsMapc = tpaps(v',v_hc',1); 
% Map all pts
markerPc = fnval(tpsMapc,markerTc'); 
%Need to clip the boundary value.
markerPc = [clipVal(markerPc(1,:),1,dim(1)); clipVal(markerPc(2,:),1,dim(2))];

% % Show the final markers. %Figure 4(b)
% figure, imshow(envsaI,[]); hold on; scatter(markerTc(:,2),markerTc(:,1),'b','filled'); 
% plot(markerPc(2,:),markerPc(1,:),'g*'); %scatter(markerP(2,:),markerP(1,:),'r');
% scatter(matBs(cind,2),matBs(cind,1),10,'r');
% title(['The final chosen landmarks.' 'ep =' num2str(ep) '; lamda =' num2str(lamda)]); hold off;

%% Create a bw mask based on the final segmentation result, use roipoly, before that do cubic spline interpolation
np = size(markerPc,2); t = linspace(0,1,np+1);
ts = linspace(0,1,np*3+1);
st = csapi(t,[markerPc(1,:) markerPc(1,1); markerPc(2,:) markerPc(2,1)]);
markerPcI = fnval(st,ts); %Interpolated.

%Remember we can scale it back
bwSegN = roipoly(envT,markerPcI(2,:)/scale,markerPcI(1,:)/scale);

% figure, hold on; scatter(markerPc(2,:),-markerPc(1,:),'filled');
% scatter(markerPcI(2,:),-markerPcI(1,:),'*'); hold off;

%% update the marker's information, so we can iterate
markerTc = markerPc';

