% Shape deformation using the bending energy and SVM regression as in Song
% Wang's paper.
%% Clean All
close all; clear all;

%% Add environment
addpath ./adam1
addpath ~/codeMatlab/nick
set(0,'DefaultTextFontSize',14);
set(0,'DefaultAxesFontSize',14);

% Read in the seismic images with segmented results, First 2D
[tmp, imgT] = readSep('imgT.H');
[hisT, envT] = readSep('envT.H');
[tmp, segT] = readSep('segT.H');

% Notice that it could be 3D.
%[hisI, imgI3] = readSep('imgI3d.H');
[hisI, envI3] = readSep('envI3d.H');
%[tmp, segI3] = readSep('segI3d.H');

%% Create the downsized version
scale = 0.5;
imgsT     = imresize(imgT,scale);
envsT     = abs(imresize(envT,scale));
segsT     = (imresize(segT,scale,'nearest'));

%% Process template image segmentation result
% Now we load the segmentation result form Adam
Iseg = (segsT)/max(segsT(:));
% Since the segmentation img values are not integers, notice that salt is
% 50.00, maximum. Therefore
thres = 1/50.0*49.00;
Im = im2bw(Iseg,thres); 
%imshow(Im); title('salt body');

% Remove the leaking tip at about (250,320) location
% [maskp, polyx, polyy] = roipoly(Im);
polyx = [229.2695 226.3973 245.1563 243.3337 242.0152 239.6711 236.6679 233.1518 229.2695];
polyy = [307.8273  324.2893  327.4158  314.1268  311.1236  310.4643  309.3655  308.3400  307.8273];
maskp = roipoly(Im, polyx, polyy);
Im = Im & ~maskp; 
%figure; imshow(Im); 

%%% The contour is rough, smooth the contour a little bit, looks like that 
%  imclose or imopen is most effective
conn = 4;
se = strel('disk',4); Io = bwareaopen(Im,200,conn); %remove the isolated small segments
Io = imfill(Io,conn,'holes'); Io = bwareaopen(Io,200,conn);
Io = imerode(Io,se); Io = bwareaopen(Io,200,conn); Io=imdilate(Io,se);
Ic = imclose(Io,strel('disk',5));

figure; imshow(Ic); %fig 3(a);

% Choose closing result, extract the contour using bwboundaries
dim = size(Im);
nx = dim(2); nz = dim(1);
mask = Ic;  % choose the Im close result
[B,L] = bwboundaries(mask,'noholes');
bw_bnd = false(dim);
bnd_coord_f = B{1};


% Now the contour coord is in bnd_coord_f, notice the order (z,x). z is downward.
% Now there is 1497 boundary pts. We down sample it to around 200 as the
% contour markers in templates.
markerDsRate = 4;
markerT = bnd_coord_f(1:markerDsRate:end,:);
%show the markers.
marker_ind = (markerT(:,2)-1)*nz + markerT(:,1);
bw_bnd(int32(marker_ind)) = true;
bw_bnd_disp = imdilate(bw_bnd,ones(2,2)); 

%%
figure; plot(markerT(:,2),markerT(:,1)); axis ij; title('template boundary'); %fig 3(b)
% figure; imshow(bw_bnd_disp); title('template marker');  %fig 3(c)

%% Init the array for entering the loop
DsegI3 = false(size(envI3)); %The array storing the deformed segmentation result
indImg = 1;
markerTc = markerT;
nmark = size(markerT,1);
selmInd = true(nmark,1);

%% Do all slices. For figure 5.
nslices = size(envI3,3);
for indImg = nslices:-1:1 %Deform the first slice only, it is the one closest to the template.
  %Create the downsized version for the input slice
  imgsI     = imresize(imgI3(:,:,indImg),scale);
  envsI     = abs(imresize(envI3(:,:,indImg),scale));
  segsI     = (imresize(segI3(:,:,indImg),scale,'nearest'));


  %Deform it using deform
  deform_R2;

  %Store the final seg mask
  DsegI3(:,:,indImg) = bwSegN;
  fprintf('processing slice %d\n', indImg); 
 
  %clips = quantile(imgsI(:),[0.03 0.97]);
  %imshow(imgsI, clips); hold on; plot(markerTc(:,2),markerTc(:,1),'b','LineWidth',2); hold off;
  %pause;
  % print out the figures using dprt
  % print(gcf,'-dpng',['oldSegOverlay_' num2str(nslices+1-indSlice)]);
end % for each input image slice

%% Write out the new segmentation result
writeSep('DsegI3d.H',hisI,double(DsegI3));

%%
plot_deformBoundary;
