function [K] = thinPlateKernel(vi,vj)
  % Calculate the thin plate kernel vector for K(v,vi).
  % Notice the kernel is U(r) = r*r*log(r), otherword, r2*0.5*log(r^2).
  % vi should be a n by 2 vector, and vj should be a m by 2 vector, and the
  % resulting vector should be a n by m vector
  n = size(vi,1); m = size(vj,1);
  assert(size(vi,2)==2 && size(vj,2)==2 );
  VIx = repmat(vi(:,1), 1,m);
  VIy = repmat(vi(:,2), 1,m);
  VJx = repmat(vj(:,1)',n,1);
  VJy = repmat(vj(:,2)',n,1);
  
  % Then subtract each coord.
  Vx = VIx - VJx;
  Vy = VIy - VJy;
  % Then calc r^2 for each (i,j) pair
  Vr = Vx(:).*Vx(:) + Vy(:).*Vy(:);
  % Calc the kernel values.
  K  = 0.5*Vr.*log(Vr);
  sel = isnan(K) | (K<1e-28); %Handle the r^2=0 case.
  K(sel) = 0;
  K  = reshape(K,[n m]);
end
